#!/usr/bin/python

import sys,os,time
from pyseq import * # (from http://www.sci.ccny.cuny.edu/~brinkman/software/pyseq/)
import threading
from numpy import * # numerical python library
from scipy.optimize import leastsq # least squares algorithm from scientific python library

# Init variables.
usbPort="/dev/ttyUSB0" # usb serial device to use
calFile="mmap.cal" # name of file for calibration data (same dir as the program)
calDumpFile="mmap.dump" # name of file for raw calib data
maxcoord=127 # max midi value to send
maxvelout=127 # max midi velocity to send
midich=0 # midi channel to use
# Modes:
MODE_STOP=0
MODE_MIDI=1
MODE_CAL=2
MODE_QUIT=3
# Other global vars.
iters=0 # number of iterations in a least squares fitting session.

#def mapCurve(p,v,s1,s2,s3,s4):
#   k1,kx,ky,kv,kx2,kxy,kxv,ky2,kyv,kv2=p
#   x=(s1+s2)-(s3+s4)
#   y=(s1+s3)-(s2+s4)
#   return k1+x*kx+y*ky+v*kv+x*x*kx2+x*y*kxy+x*v*kxv+y*y*ky2+y*v*kyv+v*v*kv2

#def mapCurve(p,v,s1,s2,s3,s4):
#   k1,k2,k3,k4,k1v,k2v,k3v,k4v=p #,k1v2,k2v2,k3v2,k4v2=p
#   return s1*k1+s2*k2+s3*k3+s4*k4+s1*v*k1v+s2*v*k2v+s3*v*k3v+s4*v*k4v #+s1*v*v*k1v2+s2*v*v*k2v2+s3*v*v*k3v2+s4*v*v*k4v2

# Defines the equation which is fitted to the raw data.
# p = array of coefficients
# v = velocity value
# s1..s4 = sensor readings
def mapCurve(p,s1,s2,s3,s4):
   #print "coeffs:", p
   k2,k3,k4,l1,l2,l3,l4=p # give names to the coefficients.
   f1=s1 # first coefficient is always 1 
   f2=s2*k2
   f3=s3*k3 
   f4=s4*k4
   v=f1+f2+f3+f4
   x=(l1*f1 +
      l2*f2 +
      l3*f3 +
      l4*f4)/v
   return x

# Puts the equation in mapCurve into a form suitable to be passed to the leastsq function.
def mapCurveErr(p,val,s1,s2,s3,s4):
   global iters
   iters+=1 # this function is called by leastsq on every iteration, so increment iters to count them.
   err=val-mapCurve(p,s1,s2,s3,s4) # leastsq needs a function which returns an error value
      # i.e. the difference between the calculated and actual x or y value.
   return err

def mapCurve2(p,x,y,v):
   a,b,c,d,e,f,g,h,i=p # give names to the coefficients.
   x=(
     a*x+
     b*y+
     c*v+
     d*x*x+
     e*y*y+
     f*v*v+
     g*x*v+
     h*y*v+
     i*x*y
     )
   return x

# Puts the equation in mapCurve into a form suitable to be passed to the leastsq function.
def mapCurve2Err(p,val,x,y,v):
   global iters
   iters+=1 # this function is called by leastsq on every iteration, so increment iters to count them.
   err=val-mapCurve2(p,x,y,v) # leastsq needs a function which returns an error value
      # i.e. the difference between the calculated and actual x or y value.
   return err


# The class which deals with reading sensor values from the arduino's serial port.
class PortReader(threading.Thread):

   # class initialisation code
   def __init__(self,usbPort):
      threading.Thread.__init__(self)
      self.usbPort=usbPort
      self.mode=MODE_STOP
      self.results=[]

   # method to return an array of results collected so far (used during calibration)
   def getResults(self):
      ret=self.results
      self.results=[]
      return ret

   # method to clear the results list
   def clearResults(self):
      self.results=[]
      return

   # the method called by the thread superclass to run the main loop of the thread.
   def run(self):
      # main loop of thread.
      print "Starting port reader."
      fhan=open(self.usbPort,'r')
      seq=PySeq() # the sequencer library object
      port=seq.createOutPort('DrumMapper') # make a midi out port.
      while True:
         line=fhan.readline().strip() # read a line from the serial port.
      #   print "abc",len(line)
         if len(line) == 0: continue
         valuesStr=line.split(", ") # turn the string of input values into an array
         values=map(lambda x: int(x), valuesStr) # convert the array to ints.
         sens=values[0:4]
         vel=sum(values); # velocity is the sum of all the sensor values.
         if self.mode == MODE_STOP: 
            print "Raw velocity:",vel,"readings:",sens
            continue
         if self.mode == MODE_MIDI:
            stage1Coords=[] # mapped coords from stage 1 calib.
            for coord in xrange(2):
               # first run the sensor readings through the first curve fit equation.
               stage1Coords.append(mapCurve(calibrator.coeffs[coord],sens[0],sens[1],sens[2],sens[3]))
            # then run the results of this through the second equation.
            midiCoords=[]
            for coord in xrange(2): # i.e. do X coord then Y
               val=mapCurve2(calibrator.coeffs2[coord],stage1Coords[0],stage1Coords[1],vel)
               val=int((val+1)*64) # convert coord into 0..127 range for midi
               # do some range checking.
               if val > 127: val=127
               if val < 0 : val=0
               midiCoords.append(val) # store the result.
            # convert velocity into midi range
            velmin,velmax=calibrator.velRange
            velOut=int(maxvelout*(vel-velmin)/(velmax-velmin))
            if velOut>maxvelout: velOut=maxvelout
            if velOut <0: velOut=0
            print "midi vel:",velOut,"; coords:",midiCoords, "; raw vel:",vel,"; readings:",sens
            # send results over midi port
            evt=snd_seq_event()
            evt.setController(midich,70,midiCoords[0])
            evt.sendNow(seq,port)
            evt=snd_seq_event()
            evt.setController(midich,71,midiCoords[1])
            evt.sendNow(seq,port)
            evt=snd_seq_event()
            evt.setNoteOn(midich,midiCoords[0],velOut)
            evt.sendNow(seq,port)
            evt=snd_seq_event()
            evt.setNoteOff(midich,midiCoords[0],velOut)
            evt.sendNow(seq,port)
            continue
         if self.mode == MODE_CAL: # in calibration mode.
            print "raw velocity:",vel, "raw sensor readings:",sens
            self.results.append(sens) # store results to be collected later.
            continue
         if self.mode == MODE_QUIT: # quit sent so end the thread.
            break

# The class which deals with calibration.
class Calibrator():
   def __init__(self):
      self.coeffs=[[],[]] # currently stored calibration coefficients
      self.coeffs2=[[],[]] # second set of coefficients.
      self.velRange=[] # range of raw velocities, min,max
      self.calLocs=[ ["top left",[1,1]], ["top centre",[1,0]], ["top right",[1,-1]], \
         ["top left mid",[0.5,0.5]], ["top right mid",[0.5,-0.5]], \
         ["left centre",[0,1]],["centre",[0,0]],["right centre",[0,-1]], \
         ["bottom left mid",[-0.5,0.5]],["bottom right mid",[-0.5,-0.5]], \
         ["bottom left",[-1,1]],["bottom centre",[-1,0]],["bottom right",[-1,-1]], \
         ] # series of locations where the user is asked to hit the pad during calibration, 
           # with names and x-y coords (on scale -1..+1 )
      self.numCalLocs=len(self.calLocs) # number of calibration locations
      self.calOK=False # is there some calibration data to work with?
      self.coeffsOK=False # are the calibration coefficients valid?
      self.inputHandle="" # file handle to read text input from.
      self.numCoeffs=7 # number of coefficients for first curve fit.
      self.numCoeffs2=9 # number of coeffs for second curve fit.
      self.rawResults=[] # stores the raw results from a calibration.
   
   # load some previously saved calibration coefficients from a file.
   def load(self):
      print "Loading calibration data... "
      calfHan=open(calFile,'r')
      self.coeffs=[0,0] # array to store x,y coefficients in.
      for coord in xrange(2): # do x coord first then y.
         line=calfHan.readline().strip()
         coordCoeffs=map(lambda x: float(x), line.split(", ")) # make the values numerical.
         self.coeffs[coord]=array(coordCoeffs) # and store them in the x,y array.
      calfHan.close()
      self.coeffsOK=True
      return

   # save calculated calibration coefficients to a file.
   def save(self):
      print "Saving calibration data... ",
      calfHan=open(calFile,'w')
      if self.coeffsOK:
         for c in self.coeffs:
            cStr=map(lambda x: str(x),c)
            cOut=""
            for i in xrange(len(cStr)):
               if i != 0: cOut += ", "
               cOut += cStr[i] 
            calfHan.write(cOut+"\n")
         calfHan.close()
         print "done"
      else:
         print "No calibration data to save"
      return

   # save actual results from sensors during calibration
   def saveRaw(self):
      print "Saving raw calibration data... ",
      calfHan=open(calDumpFile,'w')
      if self.calOK:
         for res in self.rawResults:
            locPos=res[0]
            locOut=""
            for i in xrange(2):
               if i !=0: locOut += ", "
               locOut+=str(locPos[i])
            resOut=locOut # location coords
            for sens in xrange(4):
               resOut+=", "
               resOut+=str(res[1][sens]) # a sensor reading
            calfHan.write(resOut+"\n")
         calfHan.close()
         print "done"
      else:
         print "No calibration data to save"
      return

   # load some previously saved raw calibration data.
   def loadRaw(self):
      print "Loading raw calibration data...",
      calfHan=open(calDumpFile,"r")
      self.rawResults=[]
      for line in calfHan:
         print line
         line=line.strip()
         vals=line.split(", ")
         vals=map(lambda x: float(x),vals)
         padCoords=vals[0:2]
         sens=vals[2:6]
         res=[padCoords,sens]
         print res
         self.rawResults.append(res) # put the data into rawResults as if it had just come from a calibration.
      self.calOK=True
      print "done."

   # take calibration readings from sensors in each location.
   def measure(self,inHan):
      print "Measuring calibration data"
      calResults=map(lambda x: 0.0, self.calLocs) # same number of entries as calLocs but filled with zeroes
      portReader.clearResults # clear any results from previous calibration.
      self.calOK=False # invalidate the previous calibration data.
      calLoc=0 # current calibration location (index into self.calLocs[])
      coeffsTmp=[0,0] # temp var for coefficients
      portReader.mode=MODE_CAL # switch the reader into calibration mode.
      while True:
         # prompt the user
         print "Tap the pad several times at the",self.calLocs[calLoc][0]," then type 'n'"
         print "(or 'r' to repeat a reading, or 'stop' to cancel calibration)"
         line=inHan.readline().strip()
         # interpret the command typed
         if line == "stop":
            print "Calibration stopped."
            portReader.mode=MODE_STOP
            break
         if line == "n": # go to next location.
            calResults[calLoc]=portReader.getResults() # save results from last locn.
            calLoc+=1 # move to next one.
            if calLoc==self.numCalLocs: # if it's the last one...
               # merge sensor readings from all sensors.
               self.rawResults=[]
               for loc in xrange(len(self.calLocs)): 
                  locCoords=self.calLocs[loc][1]
                  for res in calResults[loc]:
                     self.rawResults.append([locCoords,res]) # you end up with a 3D array looking like:
                        # [ [ [ x-coord , y-coord ], [ sensor1 .. sensor 4 ] ] ,
                        # ... and continue with one line for each sensor reading.
               self.calOK=True # now we have some valid data.
               break
         if line == "r": # repeat last set of readings.
            print "Please try again."
            portReader.clearResults
      print "Finished measuring calibration data."
      return

   # calculate a set of coeeficients from some recent or saved calibration data.
   def calculate(self):
      global iters
      coeffsTmp=[[],[]]
      coeffsTmp2=[[],[]]
      if not self.calOK:
         print "No Raw calibration data"
         return
      portReader.mode=MODE_STOP
      self.coeffsOK=False # invalidate the previous coefficients.
      # First work out the range of velocities in the sample data.
      print "Calculating velocity range..."
      velvals=[]
      for res in self.rawResults:
         velvals.append(sum(res[1]))
      self.velRange.append(min(velvals))
      self.velRange.append(max(velvals))
      print "   raw velocity ranges between",self.velRange[0],"and",self.velRange[1]
      print "Calculating coefficients..."
      # calculate the coefficients for the x coord then the y coord.
      mvals=[[],[]] # an array of arrays of actual x and y coords.
      for coord in xrange(2):
         print "   coordinate ",coord,":"
         sensVals=[[],[],[],[]] # an array of 1D arrays of sensor readings for each strike.
         for res in self.rawResults: # read all the raw calibration data.
            # and rearrange it into a shape suitable for the leastsq library function.
            mvals[coord].append(res[0][coord])
            for sens in xrange(4):
               sensVals[sens].append(res[1][sens]) 
         mvals[coord]=array(mvals[coord]) # convert to numpy style numerical array.
         for sens in xrange(4):
            sensVals[sens]=array(sensVals[sens]) # ditto.
         iters=0
         p0=zeros(self.numCoeffs)+1.0 # p0 is a first try at a set of coefficients for the equation.
         # now feed the rearranged results into the leastsq function.
         coeffsTmp[coord],ext=leastsq(mapCurveErr,p0,args=(mvals[coord],sensVals[0],sensVals[1],sensVals[2],sensVals[3]))
         print "      found first least squares fit in",iters,"iterations"
      # now do the second curve fit.
      # first run all the strike data through the first equation to get mapped coords for each data point.
      # start by mapping all the strike data into the right form for leastsq
      mapVals=[[],[],[]] # an array of 1D arrays of mapped x,y and v coords for each strike.
      for res in self.rawResults:
         # calculate the mapped x and y coords using 1st set of coeffs.
         for coord in xrange(2):
            mapVals[coord].append(mapCurve(coeffsTmp[coord],res[1][0],res[1][1],res[1][2],res[1][3]))
         # add the velocity coord.
         mapVals[2].append(res[1][0]+res[1][1]+res[1][2]+res[1][3])
      for coord in xrange(3):
         mapVals[coord]=array(mapVals[coord]) # turn into a numpy array.
      for coord in xrange(2): # fit x and y coords in turn
         print "   coordinate",coord,":"
         iters=0
         p0=zeros(self.numCoeffs2)+1.0 # first try at a set of coeffs.
         coeffsTmp2[coord],ext=leastsq(mapCurve2Err,p0,args=(mvals[coord],mapVals[0],mapVals[1],mapVals[2]))
         print "      found second least squares fit in",iters,"iterations"
      print "Finished calibration."
      self.coeffs=coeffsTmp # store the results
      self.coeffs2=coeffsTmp2
      print self.coeffs2
      self.coeffsOK=True # validate the calibration.
      return

# main command loop reads from stdin, accepts commands and acts on them.
def commandLoop(): 
   inhan=open("/dev/stdin",'r')
   while True:
      line=inhan.readline().strip()
      if line=="midi":
         if calibrator.coeffsOK:
            print "Now sending midi events"
            portReader.mode=MODE_MIDI
         else:
            print "Pad is not calibrated. Can't enter midi mode."
      elif line=="stop":
         print "Midi events stopped"
         portReader.mode=MODE_STOP
      elif line=="save":
         calibrator.save()
      elif line=="saveraw":
         calibrator.saveRaw()
      elif line=="calib":
         calibrator.measure(inhan)
         calibrator.calculate()
      elif line=="calfile":
         calibrator.loadRaw()
         calibrator.calculate()
      elif line=="load":
         calibrator.load()
      elif line=="quit":
         print "Quitting midimapper. Hit the pad once more to quit."
         portReader.mode=MODE_QUIT
         break
      else:
         print "No such command. Commands are midi, stop, save, saveraw, calib, load, calfile, quit"

# the main program
portReader=PortReader(usbPort) # create a thread object to read from the usb port
calibrator=Calibrator() # create a calibration object.
portReader.start() # start the read thread.
commandLoop() # start the main command loop.
portReader.join() # stop the reader thread.
