#!/usr/bin/python

import sys,os,time
from pyseq import * # (from http://www.sci.ccny.cuny.edu/~brinkman/software/pyseq/)
import threading
from numpy import * # numerical python library
from scipy.optimize import leastsq # least squares algorithm from scientific python library
import serial
import Queue
import collections

try:
 	import pygtk
  	pygtk.require("2.0")
except:
  	pass
try:
	import gtk
  	import gtk.glade
except:
	sys.exit(1)
import gobject

# configuration:
conf={'usbPort': "/dev/ttyUSB0", # usb serial device
     'usbBaud': 230400, # usb baud rate
     'calFile': "mmap.cal", # file for calibration data
     'calDumpFile': "mmap.dump", # file for raw calib data
     'velOutRange': [31,127], # range of MIDI velocities to map to
     'midiPortName': "DrumMapper", # name for alsa MIDI port
     'midich': 0, # MIDI channel to use for output
     'drawRefresh': 0.5, # drawing refresh period in seconds.
     'drawFade': 2 # time to fade strikes over in seconds. 
     }
# Constants:
#   Port Reader run modes:
MODE_STOP=0 # just read the input
MODE_OUT=1 # send output
MODE_CAL=2 # calibrate
MODE_QUIT=3 # quit
# global vars.
portReader=None # object to read from the usb serial port.
calibrator=None # object to manage calibrations
intWin=None # object to hold the GUI interface

gobject.threads_init()

# global functions
def mapCurve(p,s1,s2,s3,s4):
   """Defines the position mapping equation which is fitted to the raw data."""
   """p = array of coefficients"""
   """v = velocity value"""
   """s1..s4 = sensor readings"""
   """returns an x or y coordinate"""
   #print "coeffs:", p
   k2,k3,k4,l1,l2,l3,l4=p # give names to the coefficients.
   f1=s1 # first coefficient is always 1 
   f2=s2*k2
   f3=s3*k3 
   f4=s4*k4
   v=f1+f2+f3+f4
   x=(l1*f1 +
      l2*f2 +
      l3*f3 +
      l4*f4)/v
   return x

def mapCurveErr(self,p,val,s1,s2,s3,s4):
   """Puts the equation in mapCurve into a form suitable to be passed to the leastsq function."""
   err=val-mapCurve(p,s1,s2,s3,s4) # leastsq needs a function which returns an error value
      # i.e. the difference between the calculated and actual x or y value.
   return err

def velCurve(p,vr,x,y):
   """Curve to fit to velocity data."""
   """vr=raw velocity"""
   """x,y=mapped coordinates"""
   """returns a corrected velocity value scaled to the range [0,1]"""
   k1,k2,k3,k4,k5,k6,k7,k8,k9,k10,k11,k12=p # give names to the coefficients.
   v=k1+ k2*x+k3*y+k4*x*x+k5*y*y+k11*x*y+ \
     vr*(k6+k7*x+k8*y+k9*x*x+k10*y*y+k12*x*y)
   return v

def velCurveErr(p,vel,rawvel,x,y):
   """Error function for the velocity mapping."""
   err=vel-velCurve(p,rawvel,x,y) # leastsq needs a function which returns an error value
      # i.e. the difference between the calculated and actual velocity value.
   return err


class PortReader(threading.Thread):
   """The class which deals with reading sensor values from the arduino's serial port."""
   def __init__(self):
      """Class initialisation code."""
      threading.Thread.__init__(self)
      self.mode=MODE_STOP
      self.midiOn=True
      self.quit=False
      self.results=[]
      self.drawer=None # thread to draw strikes on the display yarea
   
   def setMode(self,mode):
      """Set the portreader's run mode if possible."""
      """Returns the mode actually set."""
      if self.mode == MODE_CAL:
         print "Can't change mode - in calibration mode."
      else:
         if mode == MODE_OUT:
            if calibrator.coeffsOK:
               self.mode=MODE_OUT
            else:
               print "Can't start output because no coefficients are loaded."
         else:
            self.mode=mode
      print "set mode to:",self.mode
      return self.mode

   def getResults(self):
      """Method to return an array of results collected so far (used during calibration)"""
      ret=self.results
      self.results=[]
      return ret

   def clearResults(self):
      """Method to clear the results list."""
      self.results=[]
      return
      
   def stop(self):
      """Quit the port reader."""
      self.quit=True
      return

   def run(self):
      """Run the main loop of the thread."""
      print "Starting port reader."
      try:
         fhan=serial.Serial(port=conf['usbPort'], baudrate=conf['usbBaud'],  timeout=0.1)
      except serial.SerialException:
         print "Can't open serial port:",conf['usbPort']
         return
      seq=PySeq() # the MIDI sequencer library object
      port=seq.createOutPort(conf['midiPortName']) # make a MIDI out port.
      midich=conf['midich']
      while True:
         inbuf=[]
         while True:
            char=fhan.read(1)
            if self.quit: 
               fhan.close()
               return
            if len(char)==0: continue
            byte=ord(char) # convert to a byte. 
            #print "%02X " % byte,
            if byte>0xff:
              print "warning - bad character" # just in case.
            inbuf.append(byte & 0x7f) # put the last 7 bits in a buffer.
            if (byte & 0x80) == 0: continue # haven't filled the buffer yet.
            else:
               # found a marker byte, so should have a full buffer.
               if len(inbuf) < 8: # buffer not full
                  print "short buffer on usb read"
                  inbuf=[]
                  continue
               else: break # have filled the buffer
         # should now have 8 * 7 bit values to make 4 * 14 bit sensor readings from.
         sens=[]
         for i in xrange(4):
            sens.append((inbuf.pop(0)<<7)+inbuf.pop(0))
         vel=sum(sens); # velocity is the sum of all the sensor values.
         # have got the sensor values, so do something with them.
         if self.mode == MODE_STOP:
            # no output or anything
            intWin.logPrintLn("raw velocity: %d; sensor values: %s" % (sum(sens),sens))
         if self.mode == MODE_OUT:
            # in output mode.
            # work out the mapped x,y coords
            xyCoords=[]
            for coord in xrange(2): # i.e. do X coord then Y
               xyVal=mapCurve(calibrator.coeffs[coord],sens[0],sens[1],sens[2],sens[3])
               if xyVal > 1.0: xyVal=1.0
               if xyVal < -1.0: xyVal=-1.0
               xyCoords.append(xyVal)
            # work out corrected velocity from velCurve
            cVel=velCurve(calibrator.velCoeffs,vel,xyCoords[0],xyCoords[1])
            if cVel > 1.0: cVel=1.0
            if cVel < 0.0: cVel=0.0
            intWin.logPrintLn("velocity: %1.2f; coords: %1.2f,%1.2f" % 
               (cVel,xyCoords[0],xyCoords[1])) # print to the log window.
            # now send something.
            if self.midiOn == True:
               midiCoords=[]
               for coord in xrange(2):
                  midiVal=int((xyCoords[coord]+1.0)*64)
                  if midiVal>127: midiVal=127
                  if midiVal<0: midiVal=0
                  midiCoords.append(midiVal)
               #print "corrected velocity:",cVel
               velOutRange=conf['velOutRange']
               velOut=int(velOutRange[0]+(cVel*(velOutRange[1]-velOutRange[0])) )
               if velOut < 0 : velOut=0
               if velOut > 127 : velOut=127
               # send results over midi port
               evt=snd_seq_event()
               evt.setController(midich,70,midiCoords[0])
               evt.sendNow(seq,port)
               evt=snd_seq_event()
               evt.setController(midich,71,midiCoords[1])
               evt.sendNow(seq,port)
               evt=snd_seq_event()
               evt.setNoteOn(midich,midiCoords[0],velOut)
               evt.sendNow(seq,port)
               evt=snd_seq_event()
               evt.setNoteOff(midich,midiCoords[0],velOut)
               evt.sendNow(seq,port)
         if self.mode == MODE_CAL: # in calibration mode.
            print "raw velocity:",vel, "raw sensor readings:",sens
            self.results.append(sens) # store results to be collected later.
            continue
         if self.mode == MODE_QUIT: # quit sent so end the thread.
            break
      return

class Calibrator():
   """The class which deals with calibration."""
   def __init__(self):
      """Initialise variables."""
      self.coeffs=[[],[]] # currently stored calibration coefficients
      self.velCoeffs=[] # range of raw velocities, min,max
      self.calLocs=[ ["top left",[1,1]], ["top centre",[1,0]], ["top right",[1,-1]], \
         ["top left mid",[0.5,0.5]], ["top right mid",[0.5,-0.5]], \
         ["left centre",[0,1]],["centre",[0,0]],["right centre",[0,-1]], \
         ["bottom left mid",[-0.5,0.5]],["bottom right mid",[-0.5,-0.5]], \
         ["bottom left",[-1,1]],["bottom centre",[-1,0]],["bottom right",[-1,-1]], \
         ] # series of locations where the user is asked to hit the pad during calibration, 
           # with names and x-y coords (on scale -1..+1 )
      self.numCalLocs=len(self.calLocs) # number of calibration locations
      self.calOK=False # is there some calibration data to work with?
      self.coeffsOK=False # are the calibration coefficients valid?
      self.inputHandle="" # file handle to read text input from.
      self.numCoeffs=7 # number of coefficients for first curve fit.
      self.numVelCoeffs=12 # number of coefficients for velocity mapping.
      self.velRange=[] # range of velocities TODO - remove this.
      self.calResults=[] # stores the raw results from a calibration.
      return
   
   def load(self):
      """Load some previously saved calibration coefficients from a file."""
      print "Loading calibration data... ",
      calfHan=open(conf['calFile'],'r')
      self.coeffs=[0,0] # array to store x,y coefficients in.
      for coord in xrange(2): # do x coord first then y.
         line=calfHan.readline().strip()
         coordCoeffs=map(lambda x: float(x), line.split(", ")) # make the values numerical.
         self.coeffs[coord]=coordCoeffs # and store them in the x,y array.
      line=calfHan.readline().strip()
      velCoeffs=map(lambda x: float(x), line.split(", "))
      self.velCoeffs=velCoeffs
      calfHan.close()
      self.coeffsOK=True
      print "done"
      return

   def save(self):
      """Save calculated calibration coefficients to a file."""
      print "Saving calibration data... ",
      calfHan=open(conf['calFile'],'w')
      if self.coeffsOK:
         for c in self.coeffs:
            cStr=map(lambda x: str(x),c)
            cOut=", ".join(cStr)
            calfHan.write(cOut+"\n")
         vStr=map(lambda x: str(x), self.velCoeffs)
         vOut=", ".join(vStr)
         calfHan.write(vOut+"\n")
         calfHan.close()
         print "done"
      else:
         print "No calibration data to save"
      return

   def saveRaw(self):
      """Save actual results from sensors during calibration."""
      print "Saving raw calibration data... ",
      calfHan=open(conf['calDumpFile'],'w')
      if self.calOK:
         for loc in xrange(len(self.calLocs)):
            locCoords=self.calLocs[loc][1]
            for res in self.calResults[loc]:
               resOut=[str(loc)]
               for sens in xrange(4):
                  resOut.append(str(res[sens]))
               strOut=", ".join(resOut)+"\n"
               calfHan.write(strOut)
         calfHan.close()
         print "Saved raw calibration data"
      else:
         print "No calibration data to save"
      return

   def loadRaw(self):
      """Load some previously saved raw calibration data."""
      print "Loading raw calibration data...",
      calfHan=open(conf['calDumpFile'],"r")
      self.calResults=[]
      for loc in xrange(len(self.calLocs)):
         self.calResults.append([])
      for line in calfHan:
         print line
         line=line.strip()
         vals=line.split(", ")
         vals=map(lambda x: int(x),vals)
         loc=vals[0] # first value is the location number.
         sens=vals[1:5] # last four are the sensor values.
         self.calResults[loc].append(sens) # put the data into calResults as if it had just come from a calibration.
      self.calOK=True
      print self.calResults
      print "done."

   def measure(self,inHan):
      """Take calibration readings from sensors in each location."""
      print "Measuring calibration data"
      self.calResults=map(lambda x: 0.0, self.calLocs) # same number of entries as calLocs but filled with zeroes
      portReader.clearResults() # clear any results from previous calibration.
      self.calOK=False # invalidate the previous calibration data.
      calLoc=0 # current calibration location (index into self.calLocs[])
      coeffsTmp=[0,0] # temp var for coefficients
      portReader.mode=MODE_CAL # switch the reader into calibration mode.
      while True:
         # prompt the user
         print "Tap the pad several times at the",self.calLocs[calLoc][0]," then type 'n'"
         print "(or 'r' to repeat a reading, or 'stop' to cancel calibration)."
         print "You should hit the pad at least 3 times each at the hardest and softest"
         print "velocity you want the pad to respond to, and a few times in between."
         line=inHan.readline().strip()
         # interpret the command typed
         if line == "stop":
            print "Calibration stopped."
            portReader.mode=MODE_STOP
            break
         if line == "n": # go to next location.
            self.calResults[calLoc]=portReader.getResults() # save results from last locn.
            portReader.clearResults()
            calLoc+=1 # move to next one.
            if calLoc==self.numCalLocs: # if it's the last one...
               self.calOK=True # now we have some valid data.
               break
         if line == "r": # repeat last set of readings.
            print "Please try again."
            portReader.clearResults()
      print "Finished measuring calibration data."
      return

   def calculate(self):
      """Calculate a set of coefficients from some recent or saved calibration data."""
      coeffsTmp=[[],[]]
      if not self.calOK:
         print "No Raw calibration data"
         return
      portReader.mode=MODE_STOP
      self.coeffsOK=False # invalidate the previous coefficients.
      print "Calculating coefficients for position mapping..."
      print "(not)"
      # calculate the coefficients for the x coord then the y coord.
      mvals=[[],[]] # an array of arrays of actual x and y coords.
      # DONE - rewritten this next bit so it merges the results from all the locations here
      # rather than when storing the data.
      for coord in xrange(2):
         print "   coordinate ",coord,":"
         sensVals=[[],[],[],[]] # an array of 1D arrays of sensor readings for each strike.
         for loc in xrange(len(self.calLocs)):
            locCoords=self.calLocs[loc][1]
            for res in self.calResults[loc]:
               mvals[coord].append(locCoords[coord])
               for sens in xrange(4):
                  sensVals[sens].append(res[sens]) 
         mvals[coord]=array(mvals[coord]) # convert to numpy style numerical array.
         for sens in xrange(4):
            sensVals[sens]=array(sensVals[sens]) # ditto.
         iters=0
         p0=zeros(self.numCoeffs)+1.0 # p0 is a first try at a set of coefficients for the equation.
         # now feed the rearranged results into the leastsq function.
         coeffsTmp[coord],cov_x,infodict,mesg,ier=leastsq(mapCurveErr,p0,args=(mvals[coord],sensVals[0],sensVals[1],sensVals[2],sensVals[3]),full_output=True,maxfev=3000)
         if ier<=4:
            print "      found least squares fit in",infodict['nfev'],"iterations"
         else:
            print "** Unable to fit curve:",mesg,"**"
            return
      print "Calculating coefficients for velocity mapping..."
      # fit the velocity mapping equation.
      cvVals=[] # an array of corrected velocities
      rvVals=[] # an array of raw velocities
      coordVals=[[],[]] # an array of two arrays of coordinate values.
      for loc in xrange(len(self.calLocs)):
         # first work out the velocity scaling for this location.
         locResults=sorted(self.calResults[loc],lambda r1,r2: cmp(sum(r1),sum(r2)))
         #print "locResults:",locResults 
         for res in xrange(len(locResults)):
            for sens in xrange(4):
               locResults[res][sens]=float(locResults[res][sens])
         velRange=[sum(locResults[0:3])/3, # ave of lowest 3 values.
                   sum(locResults[-3:])/3] # ave of highest 3 values.
         #print "velrange: ",velRange
         for res in self.calResults[loc]:
            rv=sum(res) # raw velocity is sum of sensor values.
            rvVals.append(rv) 
            cv=(float(rv)-velRange[0])/(velRange[1]-velRange[0]) # scale the corrected vel to [0,1]
            cvVals.append(cv)
            xy=[0,0]
            for coord in xrange(2):
               xy[coord]=mapCurve(coeffsTmp[coord],res[0],res[1],res[2],res[3])
               coordVals[coord].append(xy[coord])
            #print "rv:",rv,"x:",xy[0],"y:",xy[1],"cv:",cv           
      cvVals=array(cvVals)
      rvVals=array(rvVals)
      for coord in xrange(2):
         coordVals[coord]=array(coordVals[coord])
      iters=0
      p0=zeros(self.numVelCoeffs)+1.0 # p0 is a first try at a set of coefficients for the equation.
      # now feed the rearranged results into the leastsq function.
      velCoeffsTmp,cov_x,infodict,mesg,ier=leastsq(velCurveErr,p0,args=(cvVals,rvVals,coordVals[0],coordVals[1]),full_output=True)
      #print velCoeffsTmp
      if ier<=4:
         print "      found least squares fit in",infodict['nfev'],"iterations"
      else:
         print "** Unable to fit curve:",mesg,"**"
         return
      print "Finished calibration."
      self.velCoeffs=velCoeffsTmp # store the results
      self.coeffs=coeffsTmp
      self.coeffsOK=True # validate the calibration.
      return

      
class Drawer(threading.Thread):
   """The class which draws pad strikes on the display area."""
   """In a separate thread to make timing easier and reduce the load on the port reader."""
   def __init__(self):
      """Initialise the object"""
      threading.Thread.__init__(self)
      self.strikeQueue=Queue(0) # queue for passing strike data to the drawer
      fadeCycles=int(conf['drawFade']/conf['drawRefresh'])
         # cycles to fade over.
      self.fadeQueue=collections.deque([] * fadeCycles,fadeCycles) # list to store current strikes on the screen
                         # format is (v,x,y)
      self.quit=False # set this flag to end the thread.
      return
   
   def quit(self):
      """Ask the thread to end"""
      self.quit=True
      return

   def run(self):
      """Run the main loop of the thread"""
      while not self.quit:
         # first look to see if there is a new strike waiting.
         try:
            strike=self.strikeQueue.get(0) # get a strike if there is one.
            break
         except Empty:
            strike=None # if not, mark it as None
         self.fadeQueue.pop()
         self.fadeQueue.appendLeft(strike) # add it to the queue of strikes.
         # now draw all the current strikes on the screen
         print self.fadeQueue
         time.sleep(conf['drawRefresh'])
      return

class Strike:
   """Class to represent a single strike, using calibrated values."""
   def __init__(self,v,point):
      """Initialise class, with v=velocity, point=(x,y). Both should"""
      """be calibrated values."""
      self.v=v
      self.point=point
      return
      
   def __str__(self):
      """Print the strike."""
      return "(%1.2f,(%1.2f,%1.2f))" % v,point[0],point[1]

   def draw(self):
      """Draw the strike on the screen"""
      return

class InterfaceWindow:
   """Main interface window class."""
   def __init__(self):
      """Initialise the gtk interface"""
      self.gridOffset=40 # offset of the pad grid from the edge of the drawing area.
		# Set the Glade file:
      self.gladefile = "mmap.glade"
      self.wTree = gtk.glade.XML(self.gladefile, "mainWindow") 
		# Create our dictionary and connect it:
      sigDic = {"on_mainWindow_destroy" : gtk.main_quit,
		       "on_load_cal" : self.loadCal,
		       "on_save_cal": self.saveCal,
		       "on_save_raw": self.saveRaw,
		       "on_toggle_midi" : self.toggleMidi,
		       "on_toggle_output" : self.toggleOutput,
		       "on_expose_display": self.onExposeDisplay }
      self.wTree.signal_autoconnect(sigDic) 
		# Assign object variables to GUI widgets:
      widgDic = {"mainWindow" : "window",
		           "menuitem_output_on" : "outputToggleMenu",
		           "button_toggle_output" : "outputToggleButton",
		           "textview_log" : "logView",
		           "scrolledwindow_log" : "logScrollWindow",
		           "drawingarea_display" : "displayArea" }
      for (name, value) in widgDic.iteritems():
         setattr(self, value, self.wTree.get_widget(name))
      self.logBuffer = self.logView.get_buffer()
      self.vadjustLog = self.logScrollWindow.get_vadjustment()
      self.displayArea.set_size_request(500,500)
      self.makePixmap()
      # now set up the drawing area.
      self.drawLayout() 
      self.displayArea.show()
      return
      
   # Create a new backing pixmap of the appropriate size
   def makePixmap(self):
      x, y, width, height = self.displayArea.get_allocation()
      self.displayPixmap = gtk.gdk.Pixmap(self.displayArea.window, width, height)
      return
      
   def onExposeDisplay(self,widget,event):
      x , y, width, height = event.area
      self.displayArea.window.draw_drawable(self.displayArea.get_style().fg_gc[gtk.STATE_NORMAL],self.displayPixmap, x,y,x,y, width, height)
      return
   
   def _logPrint(self,str):
      """Print a string to the log window (immediate version.)"""
      self.logBuffer.insert_at_cursor(str)
      self.logView.scroll_to_mark(self.logBuffer.get_insert(),0) # scroll to bottom.
      return
   
   def logPrint(self,str):
      """Print a string to the log window. (Can be called from another thread.)"""
      gobject.idle_add(self._logPrint,str)
      return
      
   def logPrintLn(self,str):
      """Print a string plus newline to the log window. (Can be called from another thread.)"""
      gobject.idle_add(self._logPrint,str+"\n")
      return
      
   def drawLayout(self):
      """Clears the drawing window and draws the drum pad layout on it"""
      """I.e. numbers in each corner plus a grid"""
      size=list(self.displayArea.window.get_size())
      gridSize=[size[0]-2*self.gridOffset, size[1]-2*self.gridOffset]
      gridCorner=[self.gridOffset,self.gridOffset]
      self.displayPixmap.draw_rectangle(self.displayArea.get_style().white_gc,True, 0, 0, size[0],size[1]) # clear the pixmap
      colormap=self.displayArea.get_colormap()
      lightGrey=colormap.alloc_color("#aaa") # allocate a light grey color.
      greyGC=self.displayPixmap.new_gc(foreground=lightGrey)
      blackGC=self.displayArea.get_style().black_gc
      # draw the grid.
      for i in xrange(5):
         offset=( int( (gridSize[0]*i ) /4 ), int( (gridSize[1]*i )/4 ) )
         points=( [ (gridCorner[0], gridCorner[1]+offset[1] ),
                    (gridCorner[0]+gridSize[0], gridCorner[1]+offset[1]) ],
                  [ (gridCorner[0]+offset[0], gridCorner[1] ),
                    (gridCorner[0]+offset[0], gridCorner[1]+gridSize[1]) ] )
         self.displayPixmap.draw_lines(greyGC,points[0])
         self.displayPixmap.draw_lines(greyGC,points[1])         
      self.displayArea.queue_draw()
      return
      
   def loadCal(self,widget):
      """Load calibration coefficients."""
      calibrator.load()
      return
      
   def saveCal(self,widget):
      """Save calibration coefficients."""
      calibrator.save()
      return
      
   def saveRaw(self,widget):
      """Save raw calibration data."""
      calibrator.saveRaw()
      return
      
   def toggleMidi(self,widget):
      """Turn midi mode on or off."""
      portReader.midiOn=widget.get_active()
      return
      
   def toggleOutput(self,widget):
      """Turn output on or off."""
      if widget.get_active():
         mode=portReader.setMode(MODE_OUT)
      else:
         mode=portReader.setMode(MODE_STOP)
      self.outputToggleMenu.set_active(mode == MODE_OUT)
      self.outputToggleButton.set_active(mode == MODE_OUT)
      if mode==MODE_OUT:
         self.outputToggleButton.set_label("On")
      else:
         self.outputToggleButton.set_label("Off")
      return

# main program

# If the program is run directly or passed as an argument to the python
# interpreter then run the program.
if __name__ == "__main__":
   print "Starting mapper"
   portReader=PortReader()
   calibrator=Calibrator()
   intWin=InterfaceWindow()
   portReader.start()
   gtk.main() # start the gtk main loop.
   portReader.stop() # tell the port reader to quit.
   portReader.join()
   
